#include <pspkernel.h>
#include <pspdebug.h>
#include <stdlib.h>
#include <string.h>
#include <pspiofilemgr.h>
#include <pspctrl.h>
#include <pspdisplay.h>

#define O_RDONLY    0x0001
#define O_WRONLY    0x0002
#define O_RDWR      0x0003
#define O_NBLOCK    0x0010
#define O_APPEND    0x0100
#define O_CREAT     0x0200
#define O_TRUNC     0x0400
#define O_NOWAIT    0x8000

#include "colors.h"
#include "pg.h"

#define VCARD_FILE "ms0:/PSP/GAME/CONTACTS.VCF"

#define VCARD_START	"BEGIN:VCARD"
#define VCARD_END	"END:VCARD"

PSP_MODULE_INFO("vCardReader", 0, 1, 1);

int g_vcard_list_size = 0;

#define COLOR_LIST	cyan
#define COLOR_SELECTION	lawngreen

unsigned short rgb2col(unsigned char r,unsigned char g,unsigned char b)
{
    return ((((b>>3) & 0x1F)<<10)+(((g>>3) & 0x1F)<<5)+
	    (((r>>3) & 0x1F)<<0)+0x8000);
}

int exit_callback(void) 
{ 
   sceKernelExitGame(); 
   return 0; 
} 

// Thread to create the callbacks and then begin polling 
int CallbackThread(void *arg) 
{ 
   int cbid; 

   cbid = sceKernelCreateCallback("Exit Callback", exit_callback, NULL); 
   //SetExitCallback(cbid); 
   sceKernelRegisterExitCallback(cbid);

   //KernelPollCallbacks();
   sceKernelSleepThreadCB();
   return 0;
} 

/* Sets up the callback thread and returns its thread id */ 
int SetupCallbacks(void) 
{ 
   int thid = 0; 

   thid = sceKernelCreateThread("update_thread",
				CallbackThread, 0x11, 0xFA0, 0, 0);

   sceKernelStartThread(thid, 0, 0);
   
   return thid; 
} 

#define REPEAT_TIME 0x40000
static unsigned long control_bef_ctl  = 0;
static unsigned long control_bef_tick = 0;

unsigned long Read_Key(void)
{
    SceCtrlData ctl;

    sceCtrlPeekBufferPositive(&ctl, 1);
    if (ctl.Buttons == control_bef_ctl)
    {
	if ((ctl.TimeStamp - control_bef_tick) > REPEAT_TIME)
	{
	    control_bef_tick = ctl.TimeStamp;
	    return control_bef_ctl;
	}
	return 0;
    }

    control_bef_ctl  = ctl.Buttons;
    control_bef_tick = ctl.TimeStamp;

    return control_bef_ctl;
}

// like fgets but strips off \n and uses fd instead of FILE*
int file_get_string(int fd, char *buffer, int max)
{
    int i;
    char ch;

    for (i=0;i<max-1;i++)
    {
	int bytes = sceIoRead(fd, &ch, 1);

	if (bytes==0 && i == 0)
	    return -1;
	
	if (bytes == 0 || ch == '\n' || ch == '\r')
	    break;
	
	buffer[i] = ch;
    }

    buffer[i] = 0;
    return i;
}

#define FIELDCOUNT 10

enum
{
    FULLNAME,
    NAME,
    TITLE,
    ORGANIZATION,
    TELEPHONE,
    TELEPHONE_HOME,
    TELEPHONE_PREF,
    TELEPHONE_CELL,
    EMAIL,
    ADDRESS_HOME,
    ADDRESS_WORK,
};

char *fields[FIELDCOUNT] =
{
    "FN:",
    "N:",
    "TITLE:",
    "ORG:",
    "TEL:"
    "TEL;HOME:",
    "TEL;CELL:",
    "TEL;PREF:",
    "EMAIL:",
    "ADR;HOME:",
    "ADR;WORK;",
};

struct vcard
{
    char *field_data[FIELDCOUNT];
};

struct vcard_list
{
    struct vcard *card;
    struct vcard_list *next;
} g_vcard_list = {NULL, NULL};

void insert_vcard(struct vcard *card)
{
    struct vcard_list *follow_me = &g_vcard_list;

    while (follow_me->next)
    {
	follow_me = follow_me->next;
    }

    struct vcard_list *new_list_entry = malloc(sizeof(struct vcard_list));

    new_list_entry->card = card;
    new_list_entry->next = NULL;

    follow_me->next = new_list_entry;
    g_vcard_list_size ++;
}

void display_vcard_list(int x, int init_y, int start,
			int display_count, int selection)
{
    int i = 0;
    int y = init_y;
    
    struct vcard_list *the_list = g_vcard_list.next;
    
    while (the_list && display_count>0)
    {
	if (start<=0)
	{
	    if (the_list->card)
	    {
		char *display_name = NULL;

		if (the_list->card->field_data[FULLNAME])
		    display_name = the_list->card->field_data[FULLNAME];
		else if (the_list->card->field_data[ORGANIZATION])
		    display_name = the_list->card->field_data[ORGANIZATION];
		
		if (display_name)
		{
		    if (i == selection)
		    {
			pgPrint(x-2,y,magenta,"->");
			pgPrint(x,y,COLOR_SELECTION,display_name);

			draw_vcard(x+20,init_y,the_list->card);
		    }
		    else
		    {
			pgPrint(x,y,COLOR_LIST,display_name);
		    }

		    display_count--;
		    y++;
		}
	    }
	}
	else
	{
	    start--;
	}
	
	
	the_list = the_list->next;
	i++;
    }
}

void draw_vcard(int x, int y, struct vcard* this_card)
{
    int i;
    
    for (i=0;i<FIELDCOUNT;i++)
    {
	if (this_card->field_data[i])
	{
	    pgPrint(x,y,COLOR_SELECTION,fields[i]);
	    pgPrint(x+10,y,COLOR_SELECTION,this_card->field_data[i]);
	    y++;
	}
    }
}

struct vcard* new_vcard()
{
    struct vcard *this_card = malloc(sizeof(*this_card));
    int i;
    
    for (i=0;i<FIELDCOUNT;i++)
	this_card->field_data[i] = NULL;

    return this_card;
}

void insert_vcard_line(struct vcard* this_card, char *buffer)
{
    int i;
    
    for (i=0;i<FIELDCOUNT;i++)
    {
	if (strstr(buffer,fields[i]) == buffer)
	{
	    char *value = buffer + strlen(fields[i]);
	    this_card->field_data[i] = malloc(strlen(value)+1);
	    strcpy(this_card->field_data[i],value);
	    break;
	}
    }
}


struct vcard* read_next_vcard(int fd)
{
    struct vcard *this_card = new_vcard();
    
    static int y = 0;
    int i;
    char buffer[256];

    do 
    {
	int read = file_get_string(fd, buffer, 256);

	if (read<0)
	    return 0;
    } while (!strstr(buffer, VCARD_START));

    for (;;)
    {
	int read = file_get_string(fd, buffer, 256);

	if (read<0)
	    break;
	
	if (strstr(buffer, VCARD_END))
	    break;

	insert_vcard_line(this_card, buffer);
    }

    return this_card;
}



int main()
{
    SetupCallbacks();
    
    pgInit();
    pgScreenFrame(2,0);

    int fd;
    char buffer[256];

    fd = sceIoOpen(VCARD_FILE, O_RDONLY, 0777);
    
    if (!fd)
    {
	pgPrint(10,10,red,"Open Failed");
	pgScreenFlipV();
    }
    else
    {
	struct vcard *card;
	
	while ((card = read_next_vcard(fd)))
	{
	    insert_vcard(card);

	    if (card->field_data[NAME])
	    {
		pgFillvram(0);
		pgPrint(5,10,cyan,card->field_data[NAME]);
		pgScreenFlipV();
	    }
	}
    }
    
    sceIoClose(fd);

    int selection = 0;
    int start_index = 0;
    int display_count = 30;
    
    for (;;)
    {
	int keys;
	
	pgFillvram(0);

	if (selection>=start_index+display_count)
	    start_index = selection-display_count+1;

	if (selection<start_index)
	    start_index = selection;

	display_vcard_list(5, 2, start_index, display_count, selection);
	pgScreenFlipV();
	
	while ((keys = Read_Key()) == 0) ;

	if (keys & PSP_CTRL_UP)
	{
	    selection--;
	}
	if (keys & PSP_CTRL_DOWN)
	{
	    selection++;
	}
	if (keys & PSP_CTRL_LEFT)
	{
	    selection -= display_count;
	}
	if (keys & PSP_CTRL_RIGHT)
	{
	    selection += display_count;
	}
	if (keys & PSP_CTRL_RTRIGGER)
	{
	    selection = g_vcard_list_size;
	}
	if (keys & PSP_CTRL_LTRIGGER)
	{
	    selection = 0;
	}

	if (selection<0)
	    selection = 0;
	if (selection>=g_vcard_list_size-1)
	    selection = g_vcard_list_size-1;

    }
    
    return 0;
}
